/***************************************************  Programmer Functions  ***************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are used to program a flash device.

	Rev History:	05-08-2015:	Increased scope to include all four available segments in the flash.
					05-04-2015:	Initial release.

	Copyright 2015, Red Rapids, Inc.

***************************************************  Programmer Functions  ***************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "adapter.h"
#include "flash.h"
#include "common_functions.h"
#include "programmer_functions.h"

// Convert a character hex string to an unsigned long integer type.
unsigned long atoulint(char *HexNumber)
{
	unsigned long number = 0;	// Variable that will be returned by the function.
    size_t digits = 0;			// Number of digits in the string representing the unsigned integer.
    size_t loop = 0;			// Loop counter used to parse the digits in the string.
	unsigned long basemult = 1;	// Base value used to extract the number.

	// Determine the number of digits to process.
	digits = strlen(HexNumber);
	
	// Perform a hex conversion of each digit to produce the final value.
	for (loop = 0; loop < digits; loop++)
	{
		switch (HexNumber[digits-loop-1])
		{
			case '0': {number += 0 * basemult; break;}
			case '1': {number += 1 * basemult; break;}
			case '2': {number += 2 * basemult; break;}
			case '3': {number += 3 * basemult; break;}
			case '4': {number += 4 * basemult; break;}
			case '5': {number += 5 * basemult; break;}
			case '6': {number += 6 * basemult; break;}
			case '7': {number += 7 * basemult; break;}
			case '8': {number += 8 * basemult; break;}
			case '9': {number += 9 * basemult; break;}
			case 'a':
			case 'A': {number += 10 * basemult; break;}
			case 'b':
			case 'B': {number += 11 * basemult; break;}
			case 'c':
			case 'C': {number += 12 * basemult; break;}
			case 'd':
			case 'D': {number += 13 * basemult; break;}
			case 'e':
			case 'E': {number += 14 * basemult; break;}
			case 'f':
			case 'F': {number += 15 * basemult; break;}
			default: {break;}
		}
		basemult *= 16;
	}
	
	// Return the unsigned long integer result.
    return (number);
}

// Perform a blank check on the requested address range.
// Return of zero indicates blank, non-zero indicates not blank or an error condition.
unsigned long BlankCheck(s_Adapter *p_HW, unsigned long StartAddress, unsigned long EndAddress)
{
	unsigned long Address = 0;
	unsigned long Status = 0;
	
	// Set the starting address.
	Address = StartAddress;

	// Unlock all of the blocks in the address range specified.
	while (Address <= EndAddress && !Status)
	{
		// Clear the status register assigned to the current partition.
		Status += StatusClear(p_HW, (Address/PARTITION_SIZE)*PARTITION_SIZE);

		// Write the address of the block to check.
		Status += ProgAddress(p_HW, Address);

		// Issue the sequence of commands that will check the block.
		Status += ProgWrite(p_HW, BLANK_CHECK);
		Status += ProgWrite(p_HW, CONFIRM_COMMAND);

		// Wait for the blank check to complete.
		Status += StatusCheck(p_HW, (Address/PARTITION_SIZE)*PARTITION_SIZE, TYP_BLNKCHK_TIME/4);

		// Increment to the next block address.
		Address += BLOCK_SIZE;
	}
	return(Status);
}

// Erase the address range requested.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long Erase(s_Adapter *p_HW, unsigned long StartAddress, unsigned long EndAddress)
{
	unsigned long Address = 0;
	unsigned long Status = 0;
	
	// Set the starting address.
	Address = StartAddress;

	// Erase all of the blocks in the address range specified.
	while (Address <= EndAddress && !Status)
	{
		// Clear the status register assigned to the current partition.
		Status += StatusClear(p_HW, (Address/PARTITION_SIZE)*PARTITION_SIZE);

		// Write the address of the block to erase.
		Status += ProgAddress(p_HW, Address);

		// Issue the sequence of commands that will erase the block.
		Status += ProgWrite(p_HW, BLOCK_ERASE);
		Status += ProgWrite(p_HW, CONFIRM_COMMAND);

		// Wait for the erase to complete.
		Status += StatusCheck(p_HW, (Address/PARTITION_SIZE)*PARTITION_SIZE, TYP_ERASE_TIME/4);

		// Increment to the next block address.
		Address += BLOCK_SIZE;
	}
	return(Status);
}

// Report the status of the programmer clock.
// Return of zero indicates clock is stable.
unsigned long GetClockStatus(s_Adapter *p_HW)
{
	unsigned long long Value = 0;
	unsigned long Status = 0;
	
	// Save and clear the contents of the error status register.
	Status += Write(p_HW, BAR0, ERROR_STATUS, 0x0ull);
	// Read the saved copy of the error status register.
	Status += Read(p_HW, BAR0, ERROR_STATUS, &Value);

	if (Value & 0x0000000000000100ull)
		// The flash programmer clock is not locked.
		Status++;

	return(Status);
}

// Create a linked list of programming records consisting of address and 16-bit data word.
// Return NULL if an error is encountered.
s_FlashCode* MCSParser(char *MCSFileName)
{
	
	#define MAXLINE 522  // Maximum number of characters in a single line of the file.

	s_FlashCode *p_record = NULL;
	s_FlashCode *p_record_start = NULL;
	s_FlashCode *p_record_previous = NULL;
	char input_line[MAXLINE];
	char byte_count[3];
	char extended_address[5];
	char address[9];
	char data[5];
	FILE *mcsfile;
	size_t end_of_file = 0;
    unsigned long index1 = 0;
	
	// Open the MCS file for processing.
	mcsfile = fopen(MCSFileName,"r");

	// Only proceed if the file was opened successfully.
	if (mcsfile != NULL)
	{
		// Keep parsing lines of text until the end of file record is detected.
		while (!end_of_file)
		{
			// Get the next line of text.
			if (fgets(input_line,MAXLINE,mcsfile) != NULL)
			{
				// End of file record type.			
				if (input_line[8] == '1')
					// Set the end of file flag to exit the function.			
					end_of_file = 1;
				else
				{
					// Extended linear address record type.			
					if (input_line[8] == '4')
					{
						for (index1 = 0; index1 < 4; index1++)
							extended_address[index1] = input_line[9+index1];
						extended_address[4] = '\0';
					}
					else
					{
						// Data record type.			
						if (input_line[8] == '0')
						{
							// Extract the byte count.
							byte_count[0] = input_line[1];
							byte_count[1] = input_line[2];
							byte_count[2] = '\0';
					
							// Assemble the complete address.
							for (index1 = 0; index1 < 4; index1++)
								address[index1] = extended_address[index1];
							for (index1 = 4; index1 < 8; index1++)
								address[index1] = input_line[3+index1-4];
							address[8] = '\0';

							// Strip 16-bit data words from the available bytes.
							for (index1 = 0; index1 < atoulint(byte_count)/2; index1++)
							{
								// Create a new member of the linked list.
								p_record = (s_FlashCode*) malloc(sizeof(s_FlashCode));
								// Set the next pointer null since this is now the last record in the list.
								p_record->next = NULL;
								// Determine whether this is the first pointer in the linked list.
								if (p_record_start == NULL)
									// Assign the address of the first record to the start pointer.
									p_record_start = p_record;
								else
									// Assign the current address to the next pointer of the previous record.
									p_record_previous->next = p_record;
								// Store the current pointer in case another record is added to the linked list.
								p_record_previous = p_record;
						
								// Extract the current 16-bit data word.
								// Note: The data is reordered to place the least significant byte on the right.
								data[0] = input_line[9+(4*index1)+2];
								data[1] = input_line[9+(4*index1)+3];
								data[2] = input_line[9+(4*index1)];
								data[3] = input_line[9+(4*index1)+1];
								data[4] = '\0';
					
								// Store the address and data word in the current record as unsigned long values.
								p_record->address = atoulint(address)/2 + index1;
								p_record->data = atoulint(data);
							}
						}
					}
				}
			}
			// The end of file was detected before the end of file record was processed.
			else
			{
				// Set the end of file flag to exit the function.
				end_of_file = 1;
				// Free the the allocated memory.
				while (p_record_start != NULL)
				{
					// Save the location of the next record in the linked list.
					p_record = p_record_start->next;
					// Free the current record in the linked list.
					free(p_record_start);
					// Move to the next record in the linked list.
					p_record_start = p_record;
				}
			}
		}
	}
	return p_record_start;
}

// Program the address range requested.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long Program(s_Adapter *p_HW, s_FlashCode *p_FlashCode)
{
	unsigned long last_address = 0;
	unsigned long base_address = 0;
	unsigned long block_address = 0;
	unsigned long partition_address = 0;
	unsigned long Status = 0;
	s_FlashCode *p_record = NULL;
	size_t index1 = 0;
	size_t index2 = 0;
	
	// Copy the pointer to the first porgramming record.
	p_record = p_FlashCode;

	while (p_record != NULL && !Status)
	{
		// Store the base address of the current buffer.
		base_address = p_record->address;
		// Calculate the partition address of the current buffer.
		partition_address = (base_address/PARTITION_SIZE)*PARTITION_SIZE;
		// Calculate the block address of the current buffer.
		block_address = (base_address/BLOCK_SIZE)*BLOCK_SIZE;

		// Clear the status register assigned to the current partition
		Status += StatusClear(p_HW, partition_address);

		// Issue the buffered program command at the base address.
		Status += ProgAddress(p_HW, base_address);
		Status += ProgWrite(p_HW, BUFFERED_PROGRAM);

		// Load the number of words (minus one) at the block address.
		Status += ProgAddress(p_HW, block_address);
		Status += ProgWrite(p_HW, (BUFFER_SIZE-1));

		// Write an entire buffer of data in bursts that will not overflow the command FIFO.
		// Note: There are two commands (address + data) written for each program record.
		for (index1 = 0; index1 < BUFFER_SIZE/(CMD_FIFO_BURST/2); index1++)
		{
			// Wait until the command FIFO has enough room to write a complete burst.
			Status += ProgWait(p_HW);
			// Write the next burst to the command FIFO.
			// Note: There are two commands (address + data) written for each program record.
			for (index2 = 0; index2 < CMD_FIFO_BURST/2; index2++)
				// Look for the last record in the linked list.
				if (p_record != NULL)
				{
					// Load the current address and data value.
					Status += ProgAddress(p_HW, p_record->address);
					Status += ProgWrite(p_HW, p_record->data);
					// Store the current address in case this is the end of the linked list.
					last_address = p_record->address;
					// Transition to the next record in the linked list.
					p_record = p_record->next;
				}
				// Pad the remainder of the buffer with 0xFFFF if there is no more data to program.
				else
				{
					// Fill the end of the last buffer with pad data. 
					Status += ProgAddress(p_HW, ++last_address);
					Status += ProgWrite(p_HW, 0xFFFF);
				}
		}

		// Issue the confirm command.
		Status += ProgAddress(p_HW, base_address);
		Status += ProgWrite(p_HW, CONFIRM_COMMAND);

		// Wait for the programming to complete.
		Status += StatusCheck(p_HW, partition_address, TYP_BUFPROG_TIME/4);
	}
	return(Status);
}

// Load an address into the programmer.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ProgAddress(s_Adapter *p_HW, unsigned long Address)
{
	return(Write(p_HW, BAR0, BPI_BUS, ADDRESS_COMMAND | (unsigned long long)Address));
}

// Disable programming to the flash.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ProgDisable(s_Adapter *p_HW)
{
	// Disable the program voltage and chip select to the flash.
	return(Write(p_HW, BAR0, BPI_BUS, ENABLE_COMMAND));
}

// Enable the programming to the flash.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ProgEnable(s_Adapter *p_HW)
{
	// Enable the program voltage and chip select to the flash.
	return(Write(p_HW, BAR0, BPI_BUS, ENABLE_COMMAND | 0x1ul));
}

// Initialize the segment to asynchronous read mode.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ProgInit(s_Adapter *p_HW)
{
	unsigned long Status = 0;
	unsigned long index = 0;

	// Set all of the partitions in the segment to asynchronous read mode.
	for (index = 0; index < PARTITION_COUNT/SEGMENT_COUNT; index++)
	{
		// Issue the program read configuration register at every partition address.
		Status += ProgAddress(p_HW, (index*PARTITION_SIZE) | 0xD007ul);
		Status += ProgWrite(p_HW, PROG_READ_CONFIG_REG1);
		Status += ProgWrite(p_HW, PROG_READ_CONFIG_REG2);
	}

	return(Status);
}

// Perform a read operation from the programmer.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ProgRead(s_Adapter *p_HW, unsigned long *p_Value)
{
	unsigned long long Value = 0;
	unsigned long Status = 0;

	// Issue a command to read from the flash.
	Status += Write(p_HW, BAR0, BPI_BUS, READ_COMMAND);

	// Wait for a reply from the programmer.
	do
	{
		// Pause 1 usecond to reduce bus traffic.
		MicrosecondTimer(1);

		// Read the reply valid status.
		Status += Read(p_HW, BAR0, BPI_BUS, &Value);
	}
	while (!(Value & 0x0000000100000000ull));

	// Return the lower 32-bit word.
	*p_Value = (unsigned long)(Value & 0x00000000FFFFFFFFul);

	// Pop the value just read off the reply FIFO.
	Write(p_HW, BAR0, BPI_IF_CTL, 0x0ull);

	return(Status);
}

// Load the revision select into the programmer.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ProgSegment(s_Adapter *p_HW, unsigned long Revision)
{
	return(Write(p_HW, BAR0, BPI_BUS, SEGMENT_COMMAND | (unsigned long long)Revision));
}

// Wait until the command FIFO can accept another burst of data.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ProgWait(s_Adapter *p_HW)
{
	unsigned long long Value = 0;
	unsigned long Status = 0;

	// Wait for the command FIFO full flag to deassert.
	// Note: Full is programmed to allow space for a complete burst (CMD_FIFO_BURST).
	do
	{
		// Pause 5 usecond to reduce bus traffic.
		MicrosecondTimer(CMD_FIFO_BURST/4);

		// Read the command FIFO full status.
		Status += Read(p_HW, BAR0, BPI_IF_CTL, &Value);
	}
	while (Value & 0x0000000000000001ull);

	return(Status);
}

// Perform a write operation to the programmer.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ProgWrite(s_Adapter *p_HW, unsigned long Command)
{
	return(Write(p_HW, BAR0, BPI_BUS, WRITE_COMMAND | (unsigned long long)Command));
}

// Issue a command to read the ID register from the partition at the selected address.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ReadID(s_Adapter *p_HW, unsigned long Address, unsigned long Offset, unsigned long *p_Value)
{
	unsigned long Status = 0;

	// Issue the read ID command at the partition address.
	Status += ProgAddress(p_HW, Address);
	Status += ProgWrite(p_HW, READ_ID);

	// Set the offset of the desired register.
	Status += ProgAddress(p_HW, Address | Offset);

	// Read the reply from the programmer.
	Status += ProgRead(p_HW, p_Value);

	return(Status);
}

// Check the partition status bit and only return when it indicates ready.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long StatusCheck(s_Adapter *p_HW, unsigned long Address, unsigned long Timer)
{
	unsigned long Value = 0;
	unsigned long Status = 0;

	// Wait for the status to indicate that the partition is ready.
	do
	{
		// Pause to reduce bus traffic.
		MicrosecondTimer(Timer);
		// Read the status register assigned to the current partition.
		Status += StatusRead(p_HW, Address, &Value);
	}
	// Monitor the device ready bit.
	while (!(Value & 0x00000080ul));

	// Check the status for errors.
	if (Value & 0xFFFFFF7Ful)
		Status++;

	return(Status);
}

// Issue a command to clear the status register from the partition at the selected address.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long StatusClear(s_Adapter *p_HW, unsigned long Address)
{
	unsigned long Status = 0;

	// Set the partition address.
	Status += ProgAddress(p_HW, Address);

	// Issue the command to clear the register.
	ProgWrite(p_HW, CLEAR_STATUS_REGISTER);

	return(Status);
}

// Issue a command to read the status register from the partition at the selected address.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long StatusRead(s_Adapter *p_HW, unsigned long Address, unsigned long *p_Value)
{
	unsigned long Status = 0;

	// Set the partition address.
	Status += ProgAddress(p_HW, Address);

	// Read the reply from the programmer.
	Status += ProgRead(p_HW, p_Value);

	return(Status);
}

// Unlock the address range requested.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long Unlock(s_Adapter *p_HW, unsigned long StartAddress, unsigned long EndAddress)
{
	unsigned long Address = 0;
	unsigned long Value = 0;
	unsigned long Status = 0;
	
	// Set the starting address.
	Address = StartAddress;

	// Unlock all of the blocks in the address range specified.
	while (Address <= EndAddress)
	{
		// Write the address of the block to unlock.
		Status += ProgAddress(p_HW, Address);

		// Issue the sequence of commands that will unlock the block.
		Status += ProgWrite(p_HW, UNLOCK_BLOCK);
		Status += ProgWrite(p_HW, CONFIRM_COMMAND);

		// Verify that the block is unlocked.
		Status += ReadID(p_HW, (Address/PARTITION_SIZE)*PARTITION_SIZE, 0x2ul, &Value);
		// Report an error if the block is marked locked.
		if (Value & 0x00000001ul)
			Status++;

		// Increment to the next address.
		Address += BLOCK_SIZE;
	}
	return(Status);
}

// Compare the MCS file to the code loaded in flash.
// Return of zero indicates a match, non-zero indicates an error or mismatch.
unsigned long Verify(s_Adapter *p_HW, s_FlashCode *p_FlashCode)
{
	unsigned long Value = 0;
	unsigned long Status = 0;
	s_FlashCode *p_record = NULL;
	
	// Point to the first record in the linked list.
	p_record = p_FlashCode;

	// Process the entire linked list of flash code.
	while (p_record->next != NULL && !Status)
	{
		// Issue the read array command at the code address.
		Status += ProgAddress(p_HW, p_record->address);
		Status += ProgWrite(p_HW, READ_ARRAY);

		// Compare the MCS data to the value returned by the flash.
		Status += ProgRead(p_HW, &Value);
		if (p_record->data != Value)
			Status++;
		p_record = p_record->next;
	}

	return(Status);
}